use std::sync::Arc;

use crate::{
    dbus_wrappers::manager::ManagerWrapper,
    identifier::{ActiveConnectionIdentifier, ConnectionIdentifier, DeviceIdentifier},
};

use anyhow::Result;
use dbus::{nonblock::SyncConnection, Path};
use serde_derive::{Deserialize, Serialize};
use tokio::{
    io::{self, AsyncWriteExt},
    process::Command,
};

#[derive(Serialize, Deserialize, Clone, Debug)]
#[serde(tag = "type")]
pub enum Action {
    // Activates the specified connection.
    ActivateConnection {
        connection_identifier: Option<ConnectionIdentifier>,
        device_identifier: Option<DeviceIdentifier>,
        specific_object: Option<String>,
    },
    // Deactivates the specified connection.
    DeactivateConnection {
        active_connection_identifier: ActiveConnectionIdentifier,
    },
    // Executes the specified command.
    //
    // WARNING: Use caution with this option as the command will be executed
    // using the daemon's context. This could be a security issue.
    RunCommand {
        program: String,
        arguments: Vec<String>,
    },
    // Log a message to the output of the daemon. Useful for debugging.
    Log {
        message: String,
    },
}

impl Action {
    pub async fn execute(&self, conn: &Arc<SyncConnection>) -> Result<()> {
        match self {
            Action::ActivateConnection {
                connection_identifier,
                device_identifier,
                specific_object,
            } => {
                log::debug!(
                    "Executing ActivateConnection action for {:?} {:?} {:?}",
                    connection_identifier,
                    device_identifier,
                    specific_object
                );

                let maybe_connection = match connection_identifier {
                    Some(connection_identifier) => connection_identifier.to_connection(conn).await,
                    None => Ok(None),
                }?;

                let maybe_device = match device_identifier {
                    Some(device_identifier) => device_identifier.to_device(conn).await?,
                    None => None,
                };

                let manager = ManagerWrapper::from_connection(conn).await;

                let result = manager
                    .activate_connection(
                        maybe_connection.as_ref(),
                        maybe_device.as_ref(),
                        specific_object
                            .as_ref()
                            .map(|inner| Path::from(inner.clone())),
                    )
                    .await?;

                log::info!("Connection activated: {}", result.get_path());

                Ok(())
            }
            Action::DeactivateConnection {
                active_connection_identifier,
            } => {
                log::debug!(
                    "Executing DeactivateConnection action for {:?}",
                    active_connection_identifier
                );

                match active_connection_identifier
                    .to_active_connection(conn)
                    .await?
                {
                    Some(active_connection) => {
                        let manager = ManagerWrapper::from_connection(conn).await;

                        manager.deactive_connection(&active_connection).await?;
                    }
                    None => {
                        log::warn!(
                            "Unable to find active connection: {:?}",
                            active_connection_identifier
                        )
                    }
                }

                Ok(())
            }
            Action::RunCommand { program, arguments } => {
                log::debug!(
                    "Executing RunCommand action: {} {}",
                    program,
                    arguments.join(" ")
                );

                let output = Command::new(program)
                    .args(arguments.iter())
                    .output()
                    .await?;

                io::stdout().write_all(&output.stdout).await?;
                io::stderr().write_all(&output.stderr).await?;

                Ok(())
            }
            Action::Log { message } => {
                log::info!("{}", message);

                Ok(())
            }
        }
    }
}
