use std::{sync::Arc, time::Duration};

use anyhow::Result;
use dbus::{
    nonblock::{stdintf::org_freedesktop_dbus::PropertiesPropertiesChanged, Proxy, SyncConnection},
    Path,
};
use futures::{future::join_all, stream, StreamExt};
use num_traits::FromPrimitive;

use nm_reactor_dbus_codegen::network_manager::{
    OrgFreedesktopNetworkManager, OrgFreedesktopNetworkManagerDeviceAdded,
    OrgFreedesktopNetworkManagerDeviceRemoved,
};

use super::{
    active_connection::ActiveConnectionWrapper, connection::ConnectionWrapper,
    connectivity_state::ConnectivityState, device::DeviceWrapper, signal::SignalStreamWrapper,
    state::State,
};

const PATH: &str = "/org/freedesktop/NetworkManager";

pub struct ManagerWrapper<'a> {
    conn: Arc<SyncConnection>,
    inner: Box<dyn OrgFreedesktopNetworkManager + Send + Sync + 'a>,
}

impl<'a> ManagerWrapper<'a> {
    pub async fn from_connection(conn: &Arc<SyncConnection>) -> ManagerWrapper<'a> {
        let inner_conn = conn.clone();

        let proxy: Proxy<'a, Arc<SyncConnection>> = Proxy::new(
            "org.freedesktop.NetworkManager",
            PATH,
            Duration::from_millis(5000),
            inner_conn,
        );

        ManagerWrapper {
            conn: conn.clone(),
            inner: Box::new(proxy),
        }
    }

    pub async fn get_device_path_by_ip_iface(&self, iface: &str) -> anyhow::Result<Path<'_>> {
        let device_path = self.inner.get_device_by_ip_iface(iface).await?;

        Ok(device_path)
    }

    pub async fn get_all_device_paths(&self) -> anyhow::Result<Vec<Path<'a>>> {
        let device_paths = self.inner.get_all_devices().await?;

        Ok(device_paths)
    }

    pub async fn get_active_connections(
        &self,
    ) -> anyhow::Result<Vec<ActiveConnectionWrapper<'static>>> {
        let active_connection_paths = self.inner.active_connections().await?;

        let connections = active_connection_paths
            .iter()
            .map(|active_connection_path| {
                ActiveConnectionWrapper::from_path(
                    self.conn.clone(),
                    Path::from(active_connection_path.to_string()),
                )
            });

        Ok(join_all(connections).await)
    }

    pub async fn get_active_connection_from_uuid(
        &self,
        uuid: &str,
    ) -> anyhow::Result<Option<ActiveConnectionWrapper<'static>>> {
        let active_connections = self.get_active_connections().await?;

        let stream = stream::iter(active_connections);

        let mut matches: Vec<ActiveConnectionWrapper> = stream
            .filter_map(|active_connection| async move {
                let current_uuid = active_connection.get_uuid().await;

                match current_uuid {
                    Ok(current_uuid) => {
                        if current_uuid == uuid {
                            Some(active_connection)
                        } else {
                            None
                        }
                    }
                    Err(_) => None,
                }
            })
            .collect()
            .await;

        Ok(matches.pop())
    }

    pub async fn get_state(&self) -> Result<State> {
        Ok(FromPrimitive::from_u32(self.inner.state().await?).unwrap_or(State::Unknown))
    }

    pub async fn get_last_connectivity_state(&self) -> Result<ConnectivityState> {
        Ok(FromPrimitive::from_u32(self.inner.connectivity().await?)
            .unwrap_or(ConnectivityState::Unknown))
    }

    pub async fn activate_connection(
        &self,
        connection: Option<&'a ConnectionWrapper<'a>>,
        device: Option<&'a DeviceWrapper<'a>>,
        specific_object: Option<Path<'a>>,
    ) -> anyhow::Result<ActiveConnectionWrapper<'static>> {
        let active_connection_path = self
            .inner
            .activate_connection(
                connection.map_or(Path::from("/"), |c| c.get_path()),
                device.map_or(Path::from("/"), |d| d.get_path()),
                specific_object.unwrap_or_else(|| Path::from("/")),
            )
            .await?;

        let active_connection =
            ActiveConnectionWrapper::from_path(self.conn.clone(), active_connection_path).await;

        Ok(active_connection)
    }

    pub async fn deactive_connection(
        &self,
        active_connection: &ActiveConnectionWrapper<'a>,
    ) -> anyhow::Result<()> {
        self.inner
            .deactivate_connection(active_connection.get_path())
            .await?;

        Ok(())
    }

    pub async fn device_added_signal_stream(
        &self,
    ) -> anyhow::Result<SignalStreamWrapper<OrgFreedesktopNetworkManagerDeviceAdded>> {
        SignalStreamWrapper::from_match_rule(&self.conn, None, None).await
    }

    pub async fn device_removed_signal_stream(
        &self,
    ) -> anyhow::Result<SignalStreamWrapper<OrgFreedesktopNetworkManagerDeviceRemoved>> {
        SignalStreamWrapper::from_match_rule(&self.conn, None, None).await
    }

    pub async fn properties_changed_signal_stream(
        &self,
    ) -> anyhow::Result<SignalStreamWrapper<PropertiesPropertiesChanged>> {
        SignalStreamWrapper::from_match_rule(&self.conn, None, Some(Path::from(PATH))).await
    }
}
