use std::{collections::HashMap, sync::Arc};

use anyhow::{Error, Result};
use dbus::{nonblock::SyncConnection, Path};
use tokio::{
    sync::{broadcast, mpsc},
    task::JoinHandle,
};

use crate::{dbus_wrappers::manager::ManagerWrapper, device_watcher::watch_device, event::Event};

struct WatchedDevice {
    handle: tokio::task::JoinHandle<Result<(), Error>>,
    stop_signal_tx: broadcast::Sender<()>,
}

pub async fn watch(
    conn: &Arc<SyncConnection>,
    mut stop_signal_rx: broadcast::Receiver<()>,
    event_tx: mpsc::Sender<Event>,
) -> Result<JoinHandle<Result<()>>> {
    let conn2 = conn.clone();

    log::info!("Starting NetworkManager watcher task");

    let handler_handle: JoinHandle<Result<()>> = tokio::spawn(async move {
        let manager = ManagerWrapper::from_connection(&conn2).await;

        let mut watched_devices = HashMap::new();

        log::debug!("Setting up signal handlers");

        let mut device_added_signal = manager.device_added_signal_stream().await?;
        let mut device_removed_signal = manager.device_removed_signal_stream().await?;
        let mut properties_changed_signal = manager.properties_changed_signal_stream().await?;

        log::debug!("Looking for existing devices");

        for device_path in manager.get_all_device_paths().await? {
            let (stop_signal_tx, _stop_signal_rx) = broadcast::channel(1);
            let device_watcher_handle = watch_device(
                &conn2,
                stop_signal_tx.subscribe(),
                event_tx.clone(),
                device_path.clone(),
            )
            .await?;

            watched_devices.insert(
                device_path,
                WatchedDevice {
                    handle: device_watcher_handle,
                    stop_signal_tx,
                },
            );
        }

        loop {
            tokio::select! {
                Some((msg, signal)) = device_added_signal.next() => {
                    log::debug!("Got DeviceAdded signal for {}: {:?}", signal.device_path, &msg);

                    event_tx.send(Event::DeviceAdded {
                        device_path: Path::from(signal.device_path.to_string()),
                    }).await?;

                    let (stop_signal_tx, _stop_signal_rx) = broadcast::channel(1);
                    let device_watcher_handle = watch_device(&conn2, stop_signal_tx.subscribe(), event_tx.clone(), signal.device_path.clone()).await?;

                    watched_devices.insert(signal.device_path, WatchedDevice {
                        handle: device_watcher_handle,
                        stop_signal_tx
                    });
                },
                Some((msg, signal)) = device_removed_signal.next() => {
                    log::debug!("Got DeviceRemove signal for {}: {:?}", signal.device_path, &msg);

                    event_tx.send(Event::DeviceRemoved {
                        device_path: Path::from(signal.device_path.to_string()),
                    }).await?;

                    if watched_devices.contains_key(&signal.device_path) {
                        watched_devices[&signal.device_path].stop_signal_tx.send(())?;

                        watched_devices.remove(&signal.device_path);
                    }
                },
                Some((msg, signal)) = properties_changed_signal.next() => {
                    log::debug!("Got PropertiesChanged signal for {}: {:?}", signal.interface_name, &msg);

                    if signal.interface_name != "org.freedesktop.NetworkManager" {
                        break;
                    }

                    if signal.changed_properties.contains_key("Connectivity") {
                        event_tx.send(Event::ConnectivityStateChanged {
                            connectivity_state: manager.get_last_connectivity_state().await?,
                        }).await?;
                    }

                    if signal.changed_properties.contains_key("State") {
                        event_tx.send(Event::StateChanged {
                            state: manager.get_state().await?,
                        }).await?;
                    }
                }
                _ = stop_signal_rx.recv() => {
                    log::info!("Stoping NetworkManager watcher task");

                    break;
                }
            }
        }

        log::debug!("Removing signal handlers");

        device_added_signal.dispose().await?;
        device_removed_signal.dispose().await?;
        properties_changed_signal.dispose().await?;

        log::debug!("Stopping any remaining child tasks");

        for (_path, watched_device) in watched_devices.drain() {
            watched_device.stop_signal_tx.send(())?;

            watched_device.handle.await??;
        }

        Ok(())
    });

    Ok(handler_handle)
}
